"""Test Address Rule 3 formatting"""

import unittest
import paf

class TestRule3WithBuildingName(unittest.TestCase):
    """Test Address Rule 3 with Building Name Exception"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'building_name': "1A",
            'dependent_thoroughfare_name': "SEASTONE",
            'dependent_thoroughfare_descriptor': "COURT",
            'thoroughfare_name': "STATION",
            'thoroughfare_descriptor': "ROAD",
            'post_town': "HOLT",
            'postcode': "NR25 7HG"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["1A SEASTONE COURT", "STATION ROAD", "HOLT", "NR25 7HG"]
        self.assertEqual(
            self.address.as_list(), address, "Incorrect Rule 3 w/ building list format"
            )

    def test_string(self):
        """Test conversion to a string"""
        address = "1A SEASTONE COURT, STATION ROAD, HOLT. NR25 7HG"
        self.assertEqual(
            self.address.as_str(), address, "Incorrect Rule 3 w/ building string format"
            )

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("1A SEASTONE COURT", "STATION ROAD", "HOLT", "NR25 7HG")
        self.assertEqual(
            self.address.as_tuple(), address, "Incorrect Rule 3 w/ building tuple format"
            )

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "1A SEASTONE COURT",
            'line_2': "STATION ROAD",
            'post_town': "HOLT",
            'postcode': "NR25 7HG"
            }
        self.assertEqual(
            self.address.as_dict(), address, "Incorrect Rule 3 w/ building dict format"
            )

    def test_premises(self):
        """Test premises"""
        premises = {
            'premises_name': 'SEASTONE COURT',
            'sub_premises_number': 1,
            'sub_premises_suffix': 'A'
            }
        self.assertEqual(self.address.premises(), premises, "Incorrect Rule 3 w/ building premises")

class TestRule3WithoutBuildingName(unittest.TestCase):
    """Test Address Rule 3 without Building Name Exception"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'building_name': "THE MANOR",
            'thoroughfare_name': "UPPER",
            'thoroughfare_descriptor': "ROAD",
            'post_town': "HORLEY",
            'postcode': "RH6 0HP"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["THE MANOR", "UPPER ROAD", "HORLEY", "RH6 0HP"]
        self.assertEqual(
            self.address.as_list(), address, "Incorrect Rule 3 w/o building list format"
            )

    def test_string(self):
        """Test conversion to a string"""
        address = "THE MANOR, UPPER ROAD, HORLEY. RH6 0HP"
        self.assertEqual(
            self.address.as_str(), address, "Incorrect Rule 3 w/ building string format"
            )

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("THE MANOR", "UPPER ROAD", "HORLEY", "RH6 0HP")
        self.assertEqual(
            self.address.as_tuple(), address, "Incorrect Rule 3 w/o building tuple format"
            )

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "THE MANOR",
            'line_2': "UPPER ROAD",
            'post_town': "HORLEY",
            'postcode': "RH6 0HP"
            }
        self.assertEqual(
            self.address.as_dict(), address, "Incorrect Rule 3 w/o building tuple format"
            )

    def test_premises(self):
        """Test premises"""
        premises = {'premises_name': 'THE MANOR'}
        self.assertEqual(
            self.address.premises(), premises, "Incorrect Rule 3 w/o building premises"
            )

class TestRule3WithSplit(unittest.TestCase):
    """Test Address Rule 3 with Split Exception"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'organisation_name': "S D ALCOTT FLORISTS",
            'building_name': "FLOWER HOUSE 189A",
            'thoroughfare_name': "PYE GREEN",
            'thoroughfare_descriptor': "ROAD",
            'post_town': "CANNOCK",
            'postcode': "WS11 5SB"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = [
            "S D ALCOTT FLORISTS",
            "FLOWER HOUSE",
            "189A PYE GREEN ROAD",
            "CANNOCK",
            "WS11 5SB"
            ]
        self.assertEqual(self.address.as_list(), address, "Incorrect Rule 3 with split list format")

    def test_string(self):
        """Test conversion to a string"""
        address = "S D ALCOTT FLORISTS, FLOWER HOUSE, 189A PYE GREEN ROAD, CANNOCK. WS11 5SB"
        self.assertEqual(
            self.address.as_str(), address, "Incorrect Rule 3 with split string format"
            )

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = (
            "S D ALCOTT FLORISTS",
            "FLOWER HOUSE",
            "189A PYE GREEN ROAD",
            "CANNOCK",
            "WS11 5SB"
            )
        self.assertEqual(
            self.address.as_tuple(), address, "Incorrect Rule 3 with split tuple format"
            )

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "S D ALCOTT FLORISTS",
            'line_2': "FLOWER HOUSE",
            'line_3': "189A PYE GREEN ROAD",
            'post_town': "CANNOCK",
            'postcode': "WS11 5SB"
            }
        self.assertEqual(
            self.address.as_dict(), address, "Incorrect Rule 3 with split dict format"
            )

    def test_premises(self):
        """Test premises"""
        premises = {
            'premises_number': 189,
            'premises_suffix': 'A',
            'premises_name': 'FLOWER HOUSE'
            }
        self.assertEqual(self.address.premises(), premises, "Incorrect Rule 3 with split premises")

class TestRule3WithoutSplit(unittest.TestCase):
    """Test Address Rule 3 without Split Exception"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'organisation_name': "JAMES VILLA HOLIDAYS",
            'building_name': "CENTRE 30",
            'thoroughfare_name': "ST LAURENCE",
            'thoroughfare_descriptor': "AVENUE",
            'post_town': "GRAFTON",
            'postcode': "ME16 0LP"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["JAMES VILLA HOLIDAYS", "CENTRE 30", "ST LAURENCE AVENUE", "GRAFTON", "ME16 0LP"]
        self.assertEqual(self.address.as_list(), address, "Incorrect Rule 3 w/o split list format")

    def test_string(self):
        """Test conversion to a string"""
        address = "JAMES VILLA HOLIDAYS, CENTRE 30, ST LAURENCE AVENUE, GRAFTON. ME16 0LP"
        self.assertEqual(self.address.as_str(), address, "Incorrect Rule 3 w/o split string format")

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("JAMES VILLA HOLIDAYS", "CENTRE 30", "ST LAURENCE AVENUE", "GRAFTON", "ME16 0LP")
        self.assertEqual(
            self.address.as_tuple(), address, "Incorrect Rule 3 w/o split tuple format"
            )

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "JAMES VILLA HOLIDAYS",
            'line_2': "CENTRE 30",
            'line_3': "ST LAURENCE AVENUE",
            'post_town': "GRAFTON",
            'postcode': "ME16 0LP"
            }
        self.assertEqual(
            self.address.as_dict(), address, "Incorrect Rule 3 w/o split dict format"
            )

    def test_premises(self):
        """Test premises"""
        premises = {'premises_name': 'CENTRE 30'}
        self.assertEqual(self.address.premises(), premises, "Incorrect Rule 3 w/o split premises")

if __name__ == '__main__':
    unittest.main()



"""Rule 5"""

from .premisable import PremisableMixin

class Rule101(PremisableMixin):
    """Rule 5 processing"""

    @property
    def rule_attrs(self):
        """Returns premises list"""
        if getattr(self, 'concatenation_indicator', '') == 'Y':
            return ('number_sub_name_and_thoroughfare_or_locality',)
        return ('sub_building_name', 'number_and_thoroughfare_or_locality')

    @property
    def includes_first_thoroughfare_or_locality(self):
        """Returns if premises includes first thoroughfare or locality"""
        return True



"""Test Address Rule 6 formatting"""

import unittest
import paf

class TestRule6WithSubBuildingName(unittest.TestCase):
    """Test Address Rule 6 with Sub-Building Name Exception"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'sub_building_name': "10B",
            'building_name': "BARRY JACKSON TOWER",
            'thoroughfare_name': "ESTONE",
            'thoroughfare_descriptor': "WALK",
            'post_town': "BIRMINGHAM",
            'postcode': "B6 5BA"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["10B BARRY JACKSON TOWER", "ESTONE WALK", "BIRMINGHAM", "B6 5BA"]
        self.assertEqual(
            self.address.as_list(), address, "Incorrect Rule 6 with sub-building list format"
            )

    def test_string(self):
        """Test conversion to a string"""
        address = "10B BARRY JACKSON TOWER, ESTONE WALK, BIRMINGHAM. B6 5BA"
        self.assertEqual(
            self.address.as_str(), address, "Incorrect Rule 6 with sub-building string format"
            )

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("10B BARRY JACKSON TOWER", "ESTONE WALK", "BIRMINGHAM", "B6 5BA")
        self.assertEqual(
            self.address.as_tuple(), address, "Incorrect Rule 6 with sub-building tuple format"
            )

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "10B BARRY JACKSON TOWER",
            'line_2': "ESTONE WALK",
            'post_town': "BIRMINGHAM",
            'postcode': "B6 5BA"
            }
        self.assertEqual(
            self.address.as_dict(), address, "Incorrect Rule 6 with sub-building dict format"
            )

    def test_premises(self):
        """Test premises"""
        premises = {
            'premises_name': 'BARRY JACKSON TOWER',
            'sub_premises_number': 10,
            'sub_premises_suffix': 'B'
            }
        self.assertEqual(
            self.address.premises(), premises, "Incorrect Rule 6 w/ sub-building premises"
            )

class TestRule6WithBuildingName(unittest.TestCase):
    """Test Address Rule 6 with Building Name Exception"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'sub_building_name': "CARETAKERS FLAT",
            'building_name': "110-114",
            'thoroughfare_name': "HIGH",
            'thoroughfare_descriptor': "STREET WEST",
            'post_town': "BRISTOL",
            'postcode': "BS1 2AW"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["CARETAKERS FLAT", "110-114 HIGH STREET WEST", "BRISTOL", "BS1 2AW"]
        self.assertEqual(
            self.address.as_list(), address, "Incorrect Rule 6 w/ building list format"
            )

    def test_string(self):
        """Test conversion to a string"""
        address = "CARETAKERS FLAT, 110-114 HIGH STREET WEST, BRISTOL. BS1 2AW"
        self.assertEqual(
            self.address.as_str(), address, "Incorrect Rule 6 w/ building string format"
            )

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("CARETAKERS FLAT", "110-114 HIGH STREET WEST", "BRISTOL", "BS1 2AW")
        self.assertEqual(
            self.address.as_tuple(), address, "Incorrect Rule 6 w/ building tuple format"
            )

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "CARETAKERS FLAT",
            'line_2': "110-114 HIGH STREET WEST",
            'post_town': "BRISTOL",
            'postcode': "BS1 2AW"
            }
        self.assertEqual(
            self.address.as_dict(), address, "Incorrect Rule 6 w/ building dict format"
            )

    def test_premises(self):
        """Test premises"""
        premises = {
            'premises_number': 110,
            'premises_suffix': '-114',
            'sub_premises_name': 'CARETAKERS FLAT'
            }
        self.assertEqual(self.address.premises(), premises, "Incorrect Rule 6 w/ building premises")

class TestRule6(unittest.TestCase):
    """Test Address Rule 6 without Exception"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'sub_building_name': "STABLES FLAT",
            'building_name': "THE MANOR",
            'thoroughfare_name': "UPPER",
            'thoroughfare_descriptor': "HILL",
            'post_town': "HORLEY",
            'postcode': "RH6 0HP"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["STABLES FLAT", "THE MANOR", "UPPER HILL", "HORLEY", "RH6 0HP"]
        self.assertEqual(self.address.as_list(), address, "Incorrect Rule 6 list format")

    def test_string(self):
        """Test conversion to a string"""
        address = "STABLES FLAT, THE MANOR, UPPER HILL, HORLEY. RH6 0HP"
        self.assertEqual(self.address.as_str(), address, "Incorrect Rule 6 string format")

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("STABLES FLAT", "THE MANOR", "UPPER HILL", "HORLEY", "RH6 0HP")
        self.assertEqual(self.address.as_tuple(), address, "Incorrect Rule 6 tuple format")

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "STABLES FLAT",
            'line_2': "THE MANOR",
            'line_3': "UPPER HILL",
            'post_town': "HORLEY",
            'postcode': "RH6 0HP"
            }
        self.assertEqual(self.address.as_dict(), address, "Incorrect Rule 6 dict format")

    def test_premises(self):
        """Test premises"""
        premises = {'premises_name': 'THE MANOR', 'sub_premises_name': 'STABLES FLAT'}
        self.assertEqual(self.address.premises(), premises, "Incorrect Rule 6 premises")

if __name__ == '__main__':
    unittest.main()



"""Test Address with keyword arguments"""

import unittest
import paf

class TestKwargs(unittest.TestCase):
    """Test Address Keyword Arguments"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address(
            organisation_name="KATH'S CAKES",
            building_name="VICTORIA HOUSE",
            thoroughfare_name="HIGH",
            thoroughfare_descriptor="STREET",
            post_town="PORTSMOUTH",
            postcode="PO1 1AF"
            )

    def test_list(self):
        """Test conversion to an list"""
        address = [
            "KATH'S CAKES",
            "VICTORIA HOUSE",
            "HIGH STREET",
            "PORTSMOUTH",
            "PO1 1AF"
            ]
        self.assertEqual(self.address.as_list(), address, "Incorrect Kwargs list format")

    def test_string(self):
        """Test conversion to a string"""
        address = "KATH'S CAKES, VICTORIA HOUSE, HIGH STREET, PORTSMOUTH. PO1 1AF"
        self.assertEqual(self.address.as_str(), address, "Incorrect Kwargs string format")

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = (
            "KATH'S CAKES",
            "VICTORIA HOUSE",
            "HIGH STREET",
            "PORTSMOUTH",
            "PO1 1AF"
            )
        self.assertEqual(self.address.as_tuple(), address, "Incorrect Kwargs tuple format")

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "KATH'S CAKES",
            'line_2': "VICTORIA HOUSE",
            'line_3': "HIGH STREET",
            'post_town': "PORTSMOUTH",
            'postcode': "PO1 1AF"
            }
        self.assertEqual(self.address.as_dict(), address, "Incorrect Kwargs dict format")

    def test_premises(self):
        """Test premises"""
        premises = {
            'premises_name': 'VICTORIA HOUSE'
            }
        self.assertEqual(self.address.premises(), premises, "Incorrect Kwargs premises")

if __name__ == '__main__':
    unittest.main()



"""Initiator Decorator"""

import functools

def attribute_init(func):
    """Decorator to initiate an object based on a list of attributes"""

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        """Initiate properties from specified dict"""
        for key in list(getattr(self, 'attrs', [])):
            object.__setattr__(self, key, '')
        if kwargs:
            elements = kwargs
        else:
            elements = args[0]
        for key, val in elements.items():
            if hasattr(self, key):
                object.__setattr__(self, key, str(val))
        return func(self, *args, **kwargs)

    return wrapper



"""Address Premises"""

from ..initiator import attribute_init
from ..immutable import ImmutableMixin
from ..thoroughfare_locality import ThoroughfareLocalityMixin
from .attribute import AttributeMixin
from .extender import ExtenderMixin
from .lineable import LineableMixin
from .split import SplitMixin

class Premises(
    ImmutableMixin, ExtenderMixin, AttributeMixin,
        ThoroughfareLocalityMixin, LineableMixin, SplitMixin
    ):
    """PAF Address Premises class"""

    @attribute_init
    def __init__(self, *args, **kwargs): # pylint: disable=unused-argument
        """Initialise Premises elements"""
        self.extend()

    def __repr__(self):
        """Return full representation of a Premises"""
        args = {k: getattr(self, k) for k in list(self.attrs) if getattr(self, k, None)}
        return self.__class__.__name__ + '(' + str(args) + ')'

    def __str__(self):
        """Return Premises as string representation"""
        return ', '.join(self.lines)

    def __iter__(self):
        """Return Premises as iterable"""
        yield from list(self.lines)

    def __call__(self):
        """Return Premises as dictionary"""
        return self.as_premisable() # pylint: disable=no-member

    @property
    def building_number_and_sub_building_name(self):
        """Returns building number and sub-building name"""
        return self._concatenate(('building_number', 'sub_building_name'), '')

    @property
    def name_and_thoroughfare_or_locality(self):
        """Returns building number and first thoroughfare or locality"""
        return self._concatenate(('building_name', 'first_thoroughfare_or_locality'))

    @property
    def name_last_word_and_thoroughfare_or_locality(self):
        """Returns last word of building name and first thoroughfare or locality"""
        return self._concatenate(('building_name_last_word', 'first_thoroughfare_or_locality'))

    @property
    def number_and_thoroughfare_or_locality(self):
        """Returns building number and first thoroughfare or locality"""
        return self._concatenate(('building_number', 'first_thoroughfare_or_locality'))
    @property
    def number_sub_name_and_thoroughfare_or_locality(self):
        """Returns building number, sub-building name and first thoroughfare or locality"""
        return self._concatenate(
            ('building_number_and_sub_building_name', 'first_thoroughfare_or_locality')
            )

    @property
    def sub_name_and_name(self):
        """Returns sub-building name and building name"""
        return self._concatenate(('sub_building_name', 'building_name'))

    @property
    def sub_name_comma_name(self):
        """Returns sub-building name and building name concatenated with a comma"""
        return self._concatenate(('sub_building_name', 'building_name'), ', ')

    @property
    def sub_name_and_thoroughfare_or_locality(self):
        """Returns sub-building number and first thoroughfare or locality"""
        return self._concatenate(('sub_building_name', 'first_thoroughfare_or_locality'))

    @property
    def building_name_but_last_word(self):
        """Returns all but last word of the building name"""
        return self.but_last_word('building_name')

    @property
    def building_name_last_word(self):
        """Returns last word of the building name"""
        return self.last_word('building_name')

    @property
    def sub_building_name_but_last_word(self):
        """Returns all but last word of the sub-building name"""
        return self.but_last_word('sub_building_name')

    @property
    def sub_building_name_last_word(self):
        """Returns last word of the sub-building name"""
        return self.last_word('sub_building_name')

    def _concatenate(self, keys, concatenator=' '):
        """Returns specified attributes concatenated with a specified separator"""
        return concatenator.join(filter(None, [getattr(self, k, None) for k in keys]))



"""Lineable Mixin"""

from itertools import chain

class LineableMixin():
    """Converts Paf address premises elements into list of lines"""

    @property
    def lines_attrs(self):
        """Returns premises line attributes"""
        return self.organisation_attrs + ('po_box',) + self.rule_attrs

    @property
    def lines(self):
        """Returns premises lines"""
        return self._lines(self.lines_attrs)

    @property
    def po_box(self):
        """Returns PO Box"""
        return '' if self.is_empty('po_box_number') else f"PO BOX {getattr(self, 'po_box_number')}"

    def _lines(self, attrs):
        """Returns list of premises lines from specified attributes"""
        lines = list(filter(None, [getattr(self, k, None) for k in attrs]))
        return list(chain(*[line if isinstance(line, list) else [line] for line in lines]))



"""Immutable Mixin"""

from dataclasses import FrozenInstanceError

class ImmutableMixin():
    """Prevent manipulation of object attributes"""

    def __setattr__(self, *_):
        raise FrozenInstanceError

    def __delattr__(self, *_):
        raise FrozenInstanceError



"""PAF Address"""

# Tried using dataclasses.dataclass(frozen=True) decorator for immutablity but did not work"""
from .initiator import attribute_init
from .attribute import AttributeMixin
from .immutable import ImmutableMixin
from .lineable import LineableMixin
from .thoroughfare_locality import ThoroughfareLocalityMixin
from .premises.premises import Premises

class Address(ImmutableMixin, AttributeMixin, ThoroughfareLocalityMixin, LineableMixin):
    """Main PAF Address class"""

    @attribute_init
    def __init__(self, *args, **kwargs):
        """Initialise Address elements"""
        object.__setattr__(self, 'premises', Premises(*args, **kwargs))

    def __repr__(self):
        """Return full representation of an Address"""
        args = {k: getattr(self, k) for k in list(self.attrs) if getattr(self, k, None)}
        return self.__class__.__name__ + '(' + str(args) + ')'

    def __str__(self):
        """Return Address as string representation"""
        line = ', '.join(self.lines)
        if self.is_empty('postcode'):
            return line
        return '. '.join([line] + [getattr(self, 'postcode')])

    def __iter__(self):
        """Return Address as iterable"""
        yield from list(self.lines)
        if not self.is_empty('postcode'):
            yield from list([getattr(self, 'postcode')])

    def as_str(self):
        """Return Address as string"""
        return str(self)

    def as_list(self):
        """Return Address as list of strings"""
        return list(self)

    def as_tuple(self):
        """Return Address as tuple of strings"""
        return tuple(self)

    def as_dict(self):
        """Return Address as dictionary of strings"""
        address = {}
        for counter, line in enumerate(getattr(self, 'optional_lines'), 1):
            address[f"line_{counter}"] = line
        if not self.is_empty('post_town'):
            address['post_town'] = getattr(self, 'post_town')
        if not self.is_empty('postcode'):
            address['postcode'] = getattr(self, 'postcode')
        return address



"""Exceptions"""

import re
from .building_type import BuildingTypeMixin
from .split import SplitMixin

class ExceptionMixin(BuildingTypeMixin, SplitMixin):
    """Exceptions"""

    @classmethod
    def __is_exception_i(cls, val):
        """Returns if first and last characters are numeric"""
        return re.fullmatch(r'^[\d](?:.*[\d])?$', val)

    @classmethod
    def __is_exception_ii(cls, val):
        """Returns if first and penultimate characters are numeric, and last is alphabetic"""
        return re.fullmatch(r'^([\d][a-zA-Z]|[\d].*?[\d][a-zA-Z])$', val)

    @classmethod
    def __is_exception_iii(cls, val):
        """Returns if single non-whitespace character"""
        return re.fullmatch(r'^[^ \t\r\n\v\f]$', val)

    @classmethod
    def __is_exception(cls, val):
        """Returns if value is an exception"""
        return (
            cls.__is_exception_i(val)
                or cls.__is_exception_ii(val)
                or cls.__is_exception_iii(val)
            )

    def __is_exception_iv(self, attr):  # pylint: disable=unused-argument
        """Returns if value starts with a known building type
           and ends with numeric range or numeric alpha suffix"""
        # Do not include suffix check as does not account for values such as BLOCK B
        return self.is_known_building_type(attr)
        # and re.match(r'^\d', splitstring.last_word(getattr(self, attr, None)))

    def is_exception(self, attr):
        """Returns if attribute is an exception"""
        return self.__is_exception(getattr(self, attr, None))

    def is_split_exception(self, attr):
        """Returns if attribute should be split"""
        return (
            self.__is_exception(self.last_word(attr))
                and not self.last_word(attr).isdigit()
                and not self.__is_exception_iv(attr)
            )



"""Test Address Mainfile formatting"""

import unittest
import paf

class TestMainfile(unittest.TestCase):
    """Test Address Mainfile"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'organisation_name': "SOUTH LANARKSHIRE COUNCIL",
            'department_name': "HEAD START",
            'sub_building_name': "UNIT 1",
            'building_name': "BLOCK 3",
            'thoroughfare_name': "THIRD",
            'thoroughfare_descriptor': "ROAD",
            'double_dependent_locality': "BLANTYRE INDUSTRIAL ESTATE",
            'dependent_locality': "BLANTYRE",
            'post_town': "GLASGOW",
            'postcode': "G72 0UP"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = [
            "SOUTH LANARKSHIRE COUNCIL",
            "HEAD START",
            "UNIT 1",
            "BLOCK 3",
            "THIRD ROAD",
            "BLANTYRE INDUSTRIAL ESTATE",
            "BLANTYRE",
            "GLASGOW",
            "G72 0UP"
            ]
        self.assertEqual(self.address.as_list(), address, "Incorrect Mainfile list format")

    def test_string(self):
        """Test conversion to a string"""
        address = (
            "SOUTH LANARKSHIRE COUNCIL, "
            "HEAD START, "
            "UNIT 1, "
            "BLOCK 3, "
            "THIRD ROAD, "
            "BLANTYRE INDUSTRIAL ESTATE, "
            "BLANTYRE, "
            "GLASGOW. "
            "G72 0UP"
            )
        self.assertEqual(self.address.as_str(), address, "Incorrect Mainfile string format")

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = (
            "SOUTH LANARKSHIRE COUNCIL",
            "HEAD START",
            "UNIT 1",
            "BLOCK 3",
            "THIRD ROAD",
            "BLANTYRE INDUSTRIAL ESTATE",
            "BLANTYRE",
            "GLASGOW",
            "G72 0UP"
            )
        self.assertEqual(self.address.as_tuple(), address, "Incorrect Mainfile tuple format")

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "SOUTH LANARKSHIRE COUNCIL",
            'line_2': "HEAD START",
            'line_3': "UNIT 1",
            'line_4': "BLOCK 3",
            'line_5': "THIRD ROAD",
            'line_6': "BLANTYRE INDUSTRIAL ESTATE",
            'line_7': "BLANTYRE",
            'post_town': "GLASGOW",
            'postcode': "G72 0UP"
            }
        self.assertEqual(self.address.as_dict(), address, "Incorrect Mainfile dict format")

    def test_premises(self):
        """Test premises"""
        premises = {
            'premises_type': 'BLOCK',
            'premises_number': 3,
            'sub_premises_type': 'UNIT',
            'sub_premises_number': 1
            }
        self.assertEqual(self.address.premises(), premises, "Incorrect Mainfile premises")

if __name__ == '__main__':
    unittest.main()



"""Test Address Rule 2 formatting"""

import unittest
import paf

class TestRule2(unittest.TestCase):
    """Test Address Rule 2"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'building_number': "1",
            'thoroughfare_name': "ACACIA",
            'thoroughfare_descriptor': "AVENUE",
            'post_town': "ABINGDON",
            'postcode': "OX14 4PG"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["1 ACACIA AVENUE", "ABINGDON", "OX14 4PG"]
        self.assertEqual(self.address.as_list(), address, "Incorrect Rule 2 list format")

    def test_string(self):
        """Test conversion to a string"""
        address = "1 ACACIA AVENUE, ABINGDON. OX14 4PG"
        self.assertEqual(self.address.as_str(), address, "Incorrect Rule 2 string format")

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("1 ACACIA AVENUE", "ABINGDON", "OX14 4PG")
        self.assertEqual(self.address.as_tuple(), address, "Incorrect Rule 2 tuple format")

    def test_dict(self):
        """Test conversion to a dict"""
        address = {'line_1': "1 ACACIA AVENUE", 'post_town': "ABINGDON", 'postcode': "OX14 4PG"}
        self.assertEqual(self.address.as_dict(), address, "Incorrect Rule 2 dict format")

    def test_premises(self):
        """Test premises"""
        premises = {'premises_number': 1}
        self.assertEqual(self.address.premises(), premises, "Incorrect empty premises")

if __name__ == '__main__':
    unittest.main()



"""Dependent Premisable Mixin"""

from .premisable import PremisableMixin

class DependentPremisableMixin(PremisableMixin):
    """Returns the values for the premises properties"""

    @property
    def _premises_number(self):
        """Returns premises number"""
        if getattr(self, 'dependent_thoroughfare', '') != '':
            return ''
        return super()._premises_number

    @property
    def _premises_name(self):
        """Returns premises name"""
        if getattr(self, 'dependent_thoroughfare', '') != '':
            return getattr(self, 'dependent_thoroughfare', '')
        return super()._premises_name

    @property
    def _sub_premises_number(self):
        """Returns sub-premises number"""
        if getattr(self, 'dependent_thoroughfare', '') != '':
            return super()._premises_number
        return super()._sub_premises_number

    @property
    def _sub_premises_name(self):
        """Returns sub-premises name"""
        if getattr(self, 'dependent_thoroughfare', '') != '':
            return super()._premises_name
        return super()._sub_premises_name



"""Test Immutability"""

import unittest
from dataclasses import FrozenInstanceError
import paf

class TestImmutability(unittest.TestCase):
    """Test Immutable"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'building_number': "16",
            'thoroughfare_name': "VIXEN",
            'thoroughfare_descriptor': "ROAD",
            'post_town': "BRADOCK",
            'postcode': "KT6 5BT"
            })

    def test_assignment(self):
        """Test setting attribute by direct assignment"""
        with self.assertRaises(FrozenInstanceError):
            self.address.building_name = "THE HOUSE"

    def test_setattr(self):
        """Test setting new attribute"""
        self.assertRaises(FrozenInstanceError, setattr, self.address, 'building_name', "THE HOUSE")

    def test_resetattr(self):
        """Test re-setting existing attribute"""
        self.assertRaises(FrozenInstanceError, setattr, self.address, 'building_number', "20")

    def test_delattr(self):
        """Test deleting attribute"""
        self.assertRaises(FrozenInstanceError, delattr, self.address, 'post_town')

if __name__ == '__main__':
    unittest.main()



"""Lineable Mixin"""

from itertools import chain

class LineableMixin():
    """Converts Paf address elements into list of address lines"""

    @classmethod
    @property
    def optional_lines_attrs(cls):
        """Returns optional address line attributes"""
        return ('thoroughfares_and_localities',)

    @classmethod
    @property
    def lines_attrs(cls):
        """Returns optional address line attributes and post_town"""
        return cls.optional_lines_attrs + ('post_town',)

    @property
    def optional_lines(self):
        """Returns address lines, excluding post_town and postcode"""
        return self.premises.lines + self._lines(self.optional_lines_attrs)

    @property
    def lines(self):
        """Returns address lines, excluding postcode"""
        return self.premises.lines + self._lines(self.lines_attrs)

    def _lines(self, attrs):
        """Returns list of address lines from specified attributes"""
        lines = list(filter(None, [getattr(self, k, None) for k in attrs]))
        return list(chain(*[line if isinstance(line, list) else [line] for line in lines]))



"""Attribute Mixin"""

class AttributeMixin():
    """Premises elements and derived properties"""

    @classmethod
    @property
    def organisation_attrs(cls):
        """Returns Paf organisation elements"""
        return ('organisation_name', 'department_name')

    @classmethod
    @property
    def building_attrs(cls):
        """Returns Paf building elements"""
        return ('sub_building_name', 'building_name', 'building_number')

    @classmethod
    @property
    def other_attrs(cls):
        """Returns Paf other elements"""
        return ('po_box_number', 'concatenation_indicator')

    @classmethod
    @property
    def attrs(cls):
        """Returns all Paf premises elements"""
        return(
            cls.organisation_attrs
                + cls.building_attrs
                + cls.dependent_thoroughfare_attrs
                + cls.thoroughfare_attrs
                + cls.locality_attrs
                + cls.other_attrs
            )

    def is_empty(self, attr):
        """Returns if attribute value is empty"""
        return getattr(self, attr, '') == ''



"""Test Address Exception III formatting"""

import unittest
import paf

class TestExceptionIII(unittest.TestCase):
    """Test Address Exception III"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'building_name': "K",
            'thoroughfare_name': "PORTLAND",
            'thoroughfare_descriptor': "ROAD",
            'post_town': "DORKING",
            'postcode': "RH4 1EW"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["K PORTLAND ROAD", "DORKING", "RH4 1EW"]
        self.assertEqual(self.address.as_list(), address, "Incorrect Exception III list format")

    def test_string(self):
        """Test conversion to a string"""
        address = "K PORTLAND ROAD, DORKING. RH4 1EW"
        self.assertEqual(self.address.as_str(), address, "Incorrect Exception III string format")

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("K PORTLAND ROAD", "DORKING", "RH4 1EW")
        self.assertEqual(self.address.as_tuple(), address, "Incorrect Exception III tuple format")

    def test_dict(self):
        """Test conversion to a dict"""
        address = {'line_1': "K PORTLAND ROAD", 'post_town': "DORKING", 'postcode': "RH4 1EW"}
        self.assertEqual(self.address.as_dict(), address, "Incorrect Exception III dict format")

    def test_premises(self):
        """Test premises"""
        premises = {'premises_name': 'K'}
        self.assertEqual(self.address.premises(), premises, "Incorrect Exception III premises")

if __name__ == '__main__':
    unittest.main()



"""Test Address Rule 4 formatting"""

import unittest
import paf

class TestRule4(unittest.TestCase):
    """Test Address Rule 4"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'building_name': "VICTORIA HOUSE",
            'building_number': "15",
            'thoroughfare_name': "THE",
            'thoroughfare_descriptor': "STREET",
            'post_town': "CHRISTCHURCH",
            'postcode': "BH23 6AA"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["VICTORIA HOUSE", "15 THE STREET", "CHRISTCHURCH", "BH23 6AA"]
        self.assertEqual(self.address.as_list(), address, "Incorrect Rule 4 list format")

    def test_string(self):
        """Test conversion to a string"""
        address = "VICTORIA HOUSE, 15 THE STREET, CHRISTCHURCH. BH23 6AA"
        self.assertEqual(self.address.as_str(), address, "Incorrect Rule 4 string format")

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("VICTORIA HOUSE", "15 THE STREET", "CHRISTCHURCH", "BH23 6AA")
        self.assertEqual(self.address.as_tuple(), address, "Incorrect Rule 4 tuple format")

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "VICTORIA HOUSE",
            'line_2': "15 THE STREET",
            'post_town': "CHRISTCHURCH",
            'postcode': "BH23 6AA"
            }
        self.assertEqual(self.address.as_dict(), address, "Incorrect Rule 4 dict format")

    def test_premises(self):
        """Test premises"""
        premises = {'premises_number': 15, 'premises_name': 'VICTORIA HOUSE'}
        self.assertEqual(self.address.premises(), premises, "Incorrect Rule 4 premises")

if __name__ == '__main__':
    unittest.main()



"""Test Address Rule X formatting"""

import unittest
import paf

class TestRuleXWithSubBuildingName(unittest.TestCase):
    """Test Address Rule X with Sub-Building Name Exception"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'sub_building_name': "30C",
            'dependent_thoroughfare_name': "JENNENS",
            'dependent_thoroughfare_descriptor': "COURT",
            'thoroughfare_name': "JENNENS",
            'thoroughfare_descriptor': "ROAD",
            'post_town': "BIRMINGHAM",
            'postcode': "B5 5JR"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["30C JENNENS COURT", "JENNENS ROAD", "BIRMINGHAM", "B5 5JR"]
        self.assertEqual(
            self.address.as_list(), address, "Incorrect Rule X w/ sub-building list format"
            )

    def test_string(self):
        """Test conversion to a string"""
        address = "30C JENNENS COURT, JENNENS ROAD, BIRMINGHAM. B5 5JR"
        self.assertEqual(
            self.address.as_str(), address, "Incorrect Rule X w/ sub-building string format"
            )

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("30C JENNENS COURT", "JENNENS ROAD", "BIRMINGHAM", "B5 5JR")
        self.assertEqual(
            self.address.as_tuple(), address, "Incorrect Rule X w/ sub-building tuple format"
            )

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "30C JENNENS COURT",
            'line_2': "JENNENS ROAD",
            'post_town': "BIRMINGHAM",
            'postcode': "B5 5JR"
            }
        self.assertEqual(
            self.address.as_dict(), address, "Incorrect Rule X w/ sub-building dict format"
            )

    def test_premises(self):
        """Test premises"""
        premises = {
            'premises_name': 'JENNENS COURT',
            'sub_premises_number': 30,
            'sub_premises_suffix': 'C'
            }
        self.assertEqual(
            self.address.premises(), premises, "Incorrect Rule X w/ sub-building premises"
            )

class TestRuleXWithoutSubBuildingName(unittest.TestCase):
    """Test Address Rule X without Sub-Building Name Exception"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'sub_building_name': "THE ANNEX",
            'thoroughfare_name': "ST MARYS",
            'thoroughfare_descriptor': "ROAD",
            'dependent_locality': "BIRNAM",
            'post_town': "DUNKELD",
            'postcode': "PH8 0BJ"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["THE ANNEX", "ST MARYS ROAD", "BIRNAM", "DUNKELD", "PH8 0BJ"]
        self.assertEqual(
            self.address.as_list(), address, "Incorrect Rule X w/o sub-building list format"
            )

    def test_string(self):
        """Test conversion to a string"""
        address = "THE ANNEX, ST MARYS ROAD, BIRNAM, DUNKELD. PH8 0BJ"
        self.assertEqual(
            self.address.as_str(), address, "Incorrect Rule X w/o sub-building string format"
            )

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("THE ANNEX", "ST MARYS ROAD", "BIRNAM", "DUNKELD", "PH8 0BJ")
        self.assertEqual(
            self.address.as_tuple(), address, "Incorrect Rule X w/o sub-building tuple format"
            )

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "THE ANNEX",
            'line_2': "ST MARYS ROAD",
            'line_3': "BIRNAM",
            'post_town': "DUNKELD",
            'postcode': "PH8 0BJ"
            }
        self.assertEqual(
            self.address.as_dict(), address, "Incorrect Rule X w/o sub-building tuple format"
            )

    def test_premises(self):
        """Test premises"""
        premises = {'premises_name': 'THE ANNEX'}
        self.assertEqual(
            self.address.premises(), premises, "Incorrect Rule X w/o sub-building premises"
            )

if __name__ == '__main__':
    unittest.main()



"""Test Address Rule 1 formatting"""

import unittest
import paf

class TestRule1(unittest.TestCase):
    """Test Address Rule 1"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'organisation_name': "LEDA ENGINEERING LTD",
            'dependent_locality': "APPLEFORD",
            'post_town': "ABINGDON",
            'postcode': "OX14 4PG"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["LEDA ENGINEERING LTD", "APPLEFORD", "ABINGDON", "OX14 4PG"]
        self.assertEqual(self.address.as_list(), address, "Incorrect Rule 1 list format")

    def test_string(self):
        """Test conversion to a string"""
        address = "LEDA ENGINEERING LTD, APPLEFORD, ABINGDON. OX14 4PG"
        self.assertEqual(self.address.as_str(), address, "Incorrect Rule 1 string format")

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("LEDA ENGINEERING LTD", "APPLEFORD", "ABINGDON", "OX14 4PG")
        self.assertEqual(self.address.as_tuple(), address, "Incorrect Rule 1 tuple format")

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "LEDA ENGINEERING LTD",
            'line_2': "APPLEFORD",
            'post_town': "ABINGDON",
            'postcode': "OX14 4PG"
            }
        self.assertEqual(self.address.as_dict(), address, "Incorrect Rule 1 dict format")

    def test_premises(self):
        """Test premises"""
        premises = {'premises_name': 'LEDA ENGINEERING LTD'}
        self.assertEqual(self.address.premises(), premises, "Incorrect Rule 1 premises")

if __name__ == '__main__':
    unittest.main()



"""Rule X"""

from .premisable import PremisableMixin

class Rule100(PremisableMixin):
    """Rule X processing"""

    @property
    def rule_attrs(self):
        """Returns premises list"""
        if self.is_exception('sub_building_name'):
            return('sub_name_and_thoroughfare_or_locality',)
        return ('sub_building_name',)

    @property
    def includes_first_thoroughfare_or_locality(self):
        """Returns if premises includes first thoroughfare or locality"""
        return self.is_exception('sub_building_name')

    @property
    def _premises_number(self):
        """Returns premises number"""
        if getattr(self, 'dependent_thoroughfare', '') != '':
            return ''
        return super()._sub_premises_number

    @property
    def _premises_name(self):
        """Returns premises name"""
        if getattr(self, 'dependent_thoroughfare', '') != '':
            return getattr(self, 'dependent_thoroughfare', '')
        return super()._sub_premises_name

    @property
    def _sub_premises_number(self):
        """Returns sub-premises number"""
        if getattr(self, 'dependent_thoroughfare', '') != '':
            return super()._sub_premises_number
        return ''

    @property
    def _sub_premises_name(self):
        """Returns sub-premises name"""
        if getattr(self, 'dependent_thoroughfare', '') != '':
            return super()._sub_premises_name
        return ''



"""Test Address Exception II formatting"""

import unittest
import paf

class TestExceptionII(unittest.TestCase):
    """Test Address Exception II"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'building_name': "12A",
            'thoroughfare_name': "UPPERKIRKGATE",
            'post_town': "ABERDEEN",
            'postcode': "AB10 1BA"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["12A UPPERKIRKGATE", "ABERDEEN", "AB10 1BA"]
        self.assertEqual(self.address.as_list(), address, "Incorrect Exception II list format")

    def test_string(self):
        """Test conversion to a string"""
        address = "12A UPPERKIRKGATE, ABERDEEN. AB10 1BA"
        self.assertEqual(self.address.as_str(), address, "Incorrect Exception II string format")

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("12A UPPERKIRKGATE", "ABERDEEN", "AB10 1BA")
        self.assertEqual(self.address.as_tuple(), address, "Incorrect Exception II tuple format")

    def test_dict(self):
        """Test conversion to a dict"""
        address = {'line_1': "12A UPPERKIRKGATE", 'post_town': "ABERDEEN", 'postcode': "AB10 1BA"}
        self.assertEqual(self.address.as_dict(), address, "Incorrect Exception II dict format")

    def test_premises(self):
        """Test premises"""
        premises = {'premises_number': 12, 'premises_suffix': 'A'}
        self.assertEqual(self.address.premises(), premises, "Incorrect Exception II premises")

if __name__ == '__main__':
    unittest.main()



"""Test Premises attributes"""

import unittest
import paf

class TestPremisesRule1(unittest.TestCase):
    """Test Premises Rule 1"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'organisation_name': "ACME LTD",
            'dependent_thoroughfare_name': "INDUSTRIAL",
            'dependent_thoroughfare_descriptor': "ESTATE",
            'thoroughfare_name': "RODING",
            'thoroughfare_descriptor': "ROAD",
            'post_town': "LONDON",
            'postcode': "E6 7HS"
            })

    def test_premises(self):
        """Test premises"""
        premises = {'premises_name': 'INDUSTRIAL ESTATE', 'sub_premises_name': 'ACME LTD'}
        self.assertEqual(
            self.address.premises(), premises, "Incorrect Rule 1 w/ Dependent Thoroughfare"
            )

class TestPremisesRule2(unittest.TestCase):
    """Test Premises Rule 2"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'building_number': "7",
            'dependent_thoroughfare_name': "SEASTONE",
            'dependent_thoroughfare_descriptor': "COURT",
            'thoroughfare_name': "STATION",
            'thoroughfare_descriptor': "LANE",
            'post_town': "HOLT",
            'postcode': "NR25 7HG"
            })

    def test_premises(self):
        """Test premises"""
        premises = {'premises_name': 'SEASTONE COURT', 'sub_premises_number': 7}
        self.assertEqual(
            self.address.premises(), premises, "Incorrect Rule 2 w/ Dependent Thoroughfare"
            )

class TestPremisesRule4(unittest.TestCase):
    """Test Premises Rule 2"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'building_name': "CARETAKERS FLAT",
            'building_number': "99",
            'dependent_thoroughfare_name': "SEASTONE",
            'dependent_thoroughfare_descriptor': "COURT",
            'thoroughfare_name': "STATION",
            'thoroughfare_descriptor': "AVENUE",
            'post_town': "HOLT",
            'postcode': "NR25 7HG"
            })

    def test_premises(self):
        """Test premises"""
        premises = {
            'premises_name': 'SEASTONE COURT',
            'sub_premises_number': 99,
            'sub_premises_name': 'CARETAKERS FLAT'
            }
        self.assertEqual(
            self.address.premises(), premises, "Incorrect Rule 4 w/ Dependent Thoroughfare"
            )

if __name__ == '__main__':
    unittest.main()



"""Premisable Mixin"""

from .exception import ExceptionMixin

class PremisableMixin(ExceptionMixin):
    """Returns the values for the premises properties"""

    @classmethod
    @property
    def premises_attrs(cls):
        """Returns derived premises elements"""
        return (
            'premises_type', 'premises_number', 'premises_suffix', 'premises_name',
            'sub_premises_type', 'sub_premises_number', 'sub_premises_suffix', 'sub_premises_name'
            )

    @property
    def _premises_number(self):
        """Returns premises number"""
        if self.has_building_number:
            return self.building_number
        if self.is_exception('building_name') or self.is_split_exception('building_name'):
            return self.building_name_last_word
        return ''

    @property
    def _premises_name(self):
        """Returns premises name"""
        if self.has_building_number:
            return self.building_name
        if self.is_exception('building_name') or self.is_split_exception('building_name'):
            return self.building_name_but_last_word
        return self.building_name

    @property
    def _premises_name_last_word(self):
        """Returns last word of the premises name"""
        return self.last_word('_premises_name')

    @property
    def _sub_premises_number(self):
        """Returns sub-premises number"""
        if self.is_exception('sub_building_name'):
            return self.sub_building_name_last_word
        return ''

    @property
    def _sub_premises_name(self):
        """Returns sub-premises name"""
        if self.is_exception('sub_building_name'):
            return self.sub_building_name_but_last_word
        return self.sub_building_name

    @property
    def _sub_premises_name_last_word(self):
        """Returns last word of the sub-premises name"""
        return self.last_word('_sub_premises_name')

    @property
    def premises_type(self):
        """Returns premises type"""
        if self._premises_number != '':
            return ''
        if self.is_known_building_type('_premises_name'):
            return self.but_last_word('_premises_name')
        return ''

    @property
    def premises_number(self):
        """Returns premises number"""
        if self._premises_number != '':
            return self.leading_digits('_premises_number')
        if self.is_known_building_type('_premises_name'):
            return self.leading_digits('_premises_name_last_word')
        return 0

    @property
    def premises_suffix(self):
        """Returns premises suffix"""
        if self._premises_number != '':
            return self.after_leading_digits('_premises_number')
        if self.is_known_building_type('_premises_name'):
            if self.leading_digits('_premises_name_last_word') == 0:
                return self._premises_name_last_word
            return self.after_leading_digits('_premises_name_last_word')
        return ''

    @property
    def premises_name(self):
        """Returns premises name"""
        if self._premises_number != '':
            if self.leading_digits('_premises_number') == 0:
                return self._premises_number
            return self._premises_name
        if self.is_known_building_type('_premises_name'):
            return ''
        return self._premises_name

    @property
    def sub_premises_type(self):
        """Returns sub-premises type"""
        if self._sub_premises_number != '':
            return ''
        if self.is_known_sub_building_type('_sub_premises_name'):
            return self.but_last_word('_sub_premises_name')
        return ''

    @property
    def sub_premises_number(self):
        """Returns sub-premises number"""
        if self._sub_premises_number != '':
            return self.leading_digits('_sub_premises_number')
        if self.is_known_sub_building_type('_sub_premises_name'):
            return self.leading_digits('_sub_premises_name_last_word')
        return 0

    @property
    def sub_premises_suffix(self):
        """Returns sub-premises suffix"""
        if self._sub_premises_number != '':
            return self.after_leading_digits('_sub_premises_number')
        if self.is_known_sub_building_type('_sub_premises_name'):
            if self.leading_digits('_sub_premises_name_last_word') == 0:
                return self._sub_premises_name_last_word
            return self.after_leading_digits('_sub_premises_name_last_word')
        return ''

    @property
    def sub_premises_name(self):
        """Returns sub-premises name"""
        if self._sub_premises_number != '':
            if self.leading_digits('_sub_premises_number') == 0:
                return self._sub_premises_number
            return self._sub_premises_name
        if self.is_known_sub_building_type('_sub_premises_name'):
            return ''
        return self._sub_premises_name

    @property
    def has_building_number(self):
        """Returns if address has a populated building number"""
        return self.building_number not in ('', '0')

    def as_premisable(self):
        """Return Premisable as dictionary"""
        return {
            attr: getattr(self, attr) for attr in self.premises_attrs
            if getattr(self, attr, None) and getattr(self, attr) != 0
            }



"""Rule 1"""

from .premisable import PremisableMixin

class Rule000(PremisableMixin):
    """Rule 1 processing"""

    @property
    def rule_attrs(self):
        """Returns premises list"""
        return ()

    @property
    def includes_first_thoroughfare_or_locality(self):
        """Returns if premises includes first thoroughfare or locality"""
        return False

    @property
    def _premises_name(self):
        """Returns premises name"""
        if getattr(self, 'dependent_thoroughfare', '') != '':
            return getattr(self, 'dependent_thoroughfare', '')
        if getattr(self, 'organisation_name', '') != '':
            return getattr(self, 'organisation_name', '')
        if getattr(self, 'po_box', '') != '':
            return getattr(self, 'po_box', '')
        return super()._premises_name

    @property
    def _sub_premises_name(self):
        """Returns sub-premises name"""
        if getattr(self, 'dependent_thoroughfare', '') != '':
            if getattr(self, 'organisation_name', '') != '':
                return getattr(self, 'organisation_name', '')
            if getattr(self, 'po_box', '') != '':
                return getattr(self, 'po_box', '')
        return super()._sub_premises_name



"""Test Address Exception I formatting"""

import unittest
import paf

class TestExceptionI(unittest.TestCase):
    """Test Address Exception I"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'building_name': "1-2",
            'thoroughfare_name': "NURSERY",
            'thoroughfare_descriptor': "LANE",
            'dependent_locality': "PENN",
            'post_town': "HIGH WYCOMBE",
            'postcode': "HP10 8LS"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["1-2 NURSERY LANE", "PENN", "HIGH WYCOMBE", "HP10 8LS"]
        self.assertEqual(self.address.as_list(), address, "Incorrect Exception I list format")

    def test_string(self):
        """Test conversion to a string"""
        address = "1-2 NURSERY LANE, PENN, HIGH WYCOMBE. HP10 8LS"
        self.assertEqual(self.address.as_str(), address, "Incorrect Exception I string format")

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("1-2 NURSERY LANE", "PENN", "HIGH WYCOMBE", "HP10 8LS")
        self.assertEqual(self.address.as_tuple(), address, "Incorrect Exception I tuple format")

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "1-2 NURSERY LANE",
            'line_2': "PENN",
            'post_town': "HIGH WYCOMBE",
            'postcode': "HP10 8LS"
            }
        self.assertEqual(self.address.as_dict(), address, "Incorrect Exception I dict format")

    def test_premises(self):
        """Test premises"""
        premises = {'premises_number': 1, 'premises_suffix': '-2'}
        self.assertEqual(self.address.premises(), premises, "Incorrect Exception I premises")

if __name__ == '__main__':
    unittest.main()



"""Test Address Rule 7 formatting"""

import unittest
import paf

class TestRule7WithZeroBuildingNumber(unittest.TestCase):
    """Test Address Rule 7 with a 0 Building Number"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'sub_building_name': "FLAT 1",
            'building_name': "HOLLY HOUSE",
            'building_number': "0",
            'thoroughfare_name': "OAK",
            'thoroughfare_descriptor': "AVENUE",
            'dependent_locality': "BIDDENDEN",
            'post_town': "ASHFORD",
            'postcode': "TN27 8BT"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["FLAT 1, HOLLY HOUSE", "OAK AVENUE", "BIDDENDEN", "ASHFORD", "TN27 8BT"]
        self.assertEqual(
            self.address.as_list(), address, "Incorrect Rule 7 with 0 building number list format"
            )

    def test_string(self):
        """Test conversion to a string"""
        address = "FLAT 1, HOLLY HOUSE, OAK AVENUE, BIDDENDEN, ASHFORD. TN27 8BT"
        self.assertEqual(
            self.address.as_str(), address, "Incorrect Rule 7 with 0 building number string format"
            )

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("FLAT 1, HOLLY HOUSE", "OAK AVENUE", "BIDDENDEN", "ASHFORD", "TN27 8BT")
        self.assertEqual(
            self.address.as_tuple(), address, "Incorrect Rule 7 with 0 building number tuple format"
            )

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "FLAT 1, HOLLY HOUSE",
            'line_2': "OAK AVENUE",
            'line_3': "BIDDENDEN",
            'post_town': "ASHFORD",
            'postcode': "TN27 8BT"
            }
        self.assertEqual(
            self.address.as_dict(), address, "Incorrect Rule 7 with 0 building number dict format"
            )

    def test_premises(self):
        """Test premises"""
        premises = {
            'premises_name': 'HOLLY HOUSE',
            'sub_premises_type': 'FLAT',
            'sub_premises_number': 1
            }
        self.assertEqual(
            self.address.premises(), premises, "Incorrect Rule 7 with 0 building number premises"
            )

class TestRule7WithSubBuildingName(unittest.TestCase):
    """Test Address Rule 7 with Sub-Building Name Exception"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'sub_building_name': "2B",
            'building_name': "THE TOWER",
            'building_number': "27",
            'thoroughfare_name': "JOHN",
            'thoroughfare_descriptor': "STREET",
            'post_town': "WINCHESTER",
            'postcode': "SO23 9AP"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["2B THE TOWER", "27 JOHN STREET", "WINCHESTER", "SO23 9AP"]
        self.assertEqual(
            self.address.as_list(), address, "Incorrect Rule 7 with sub-building list format"
            )

    def test_string(self):
        """Test conversion to a string"""
        address = "2B THE TOWER, 27 JOHN STREET, WINCHESTER. SO23 9AP"
        self.assertEqual(
            self.address.as_str(), address, "Incorrect Rule 7 with sub-building string format"
            )

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("2B THE TOWER", "27 JOHN STREET", "WINCHESTER", "SO23 9AP")
        self.assertEqual(
            self.address.as_tuple(), address, "Incorrect Rule 7 with sub-building tuple format"
            )

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "2B THE TOWER",
            'line_2': "27 JOHN STREET",
            'post_town': "WINCHESTER",
            'postcode': "SO23 9AP"
            }
        self.assertEqual(
            self.address.as_dict(), address, "Incorrect Rule 7 with sub-building dict format"
            )

    def test_premises(self):
        """Test premises"""
        premises = {
            'premises_number': 27,
            'premises_name': 'THE TOWER',
            'sub_premises_number': 2,
            'sub_premises_suffix': 'B'
            }
        self.assertEqual(
            self.address.premises(), premises, "Incorrect Rule 7 with sub-building premises"
            )

class TestRule7(unittest.TestCase):
    """Test Address Rule 7 without Exception"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'sub_building_name': "BASEMENT FLAT",
            'building_name': "VICTORIA HOUSE",
            'building_number': "15",
            'thoroughfare_name': "THE",
            'thoroughfare_descriptor': "STREET",
            'post_town': "CORYTON",
            'postcode': "BP23 6AA"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["BASEMENT FLAT", "VICTORIA HOUSE", "15 THE STREET", "CORYTON", "BP23 6AA"]
        self.assertEqual(self.address.as_list(), address, "Incorrect Rule 7 list format")

    def test_string(self):
        """Test conversion to a string"""
        address = "BASEMENT FLAT, VICTORIA HOUSE, 15 THE STREET, CORYTON. BP23 6AA"
        self.assertEqual(self.address.as_str(), address, "Incorrect Rule 7 string format")

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("BASEMENT FLAT", "VICTORIA HOUSE", "15 THE STREET", "CORYTON", "BP23 6AA")
        self.assertEqual(self.address.as_tuple(), address, "Incorrect Rule 7 tuple format")

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "BASEMENT FLAT",
            'line_2': "VICTORIA HOUSE",
            'line_3': "15 THE STREET",
            'post_town': "CORYTON",
            'postcode': "BP23 6AA"
            }
        self.assertEqual(self.address.as_dict(), address, "Incorrect Rule 7 dict format")

    def test_premises(self):
        """Test premises"""
        premises = {
            'premises_number': 15,
            'premises_name': 'VICTORIA HOUSE',
            'sub_premises_name': 'BASEMENT FLAT'
            }
        self.assertEqual(self.address.premises(), premises, "Incorrect Rule 7 premises")

if __name__ == '__main__':
    unittest.main()



"""Rule 2"""

from .dependent_premisable import DependentPremisableMixin

class Rule001(DependentPremisableMixin):
    """Rule 2 processing"""

    @property
    def rule_attrs(self):
        """Returns premises list"""
        return ('number_and_thoroughfare_or_locality',)

    @property
    def includes_first_thoroughfare_or_locality(self):
        """Returns if premises includes first thoroughfare or locality"""
        return True



"""Test Empty Address formatting"""

import unittest
import paf

class TestEmpty(unittest.TestCase):
    """Test Address Exception I"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({})

    def test_list(self):
        """Test conversion to an list"""
        address = []
        self.assertEqual(self.address.as_list(), address, "Incorrect empty list format")

    def test_string(self):
        """Test conversion to a string"""
        address = ""
        self.assertEqual(self.address.as_str(), address, "Incorrect empty string format")

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ()
        self.assertEqual(self.address.as_tuple(), address, "Incorrect empty tuple format")

    def test_dict(self):
        """Test conversion to a dict"""
        address = {}
        self.assertEqual(self.address.as_dict(), address, "Incorrect empty dict format")

    def test_premises(self):
        """Test premises"""
        premises = {}
        self.assertEqual(self.address.premises(), premises, "Incorrect empty premises")

if __name__ == '__main__':
    unittest.main()



"""Rule 4"""

from .dependent_premisable import DependentPremisableMixin

class Rule011(DependentPremisableMixin):
    """Rule 4 processing"""

    @property
    def rule_attrs(self):
        """Returns premises list"""
        return ('building_name', 'number_and_thoroughfare_or_locality')

    @property
    def includes_first_thoroughfare_or_locality(self):
        """Returns if premises includes first thoroughfare or locality"""
        return True



"""Test Address Exception IV formatting"""

import unittest
import paf

class TestExceptionIVUnit(unittest.TestCase):
    """Test Address Exception IV Unit"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'organisation_name': "THE TAMBOURINE WAREHOUSE",
            'building_name': "UNIT 1-3",
            'dependent_thoroughfare_name': "INDUSTRIAL",
            'dependent_thoroughfare_descriptor': "ESTATE",
            'thoroughfare_name': "TAME",
            'thoroughfare_descriptor': "ROAD",
            'post_town': "LONDON",
            'postcode': "E6 7HS"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = [
            "THE TAMBOURINE WAREHOUSE",
            "UNIT 1-3",
            "INDUSTRIAL ESTATE",
            "TAME ROAD",
            "LONDON",
            "E6 7HS"
            ]
        self.assertEqual(self.address.as_list(), address, "Incorrect Exception IV Unit list format")

    def test_string(self):
        """Test conversion to a string"""
        address = "THE TAMBOURINE WAREHOUSE, UNIT 1-3, INDUSTRIAL ESTATE, TAME ROAD, LONDON. E6 7HS"
        self.assertEqual(
            self.address.as_str(), address, "Incorrect Exception IV Unit string format"
            )

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = (
            "THE TAMBOURINE WAREHOUSE",
            "UNIT 1-3",
            "INDUSTRIAL ESTATE",
            "TAME ROAD",
            "LONDON",
            "E6 7HS"
            )
        self.assertEqual(
            self.address.as_tuple(), address, "Incorrect Exception IV Unit tuple format"
            )

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "THE TAMBOURINE WAREHOUSE",
            'line_2': "UNIT 1-3",
            'line_3': "INDUSTRIAL ESTATE",
            'line_4': "TAME ROAD",
            'post_town': "LONDON",
            'postcode': "E6 7HS"
            }
        self.assertEqual(
            self.address.as_dict(), address, "Incorrect Exception IV Unit dict format"
            )

    def test_premises(self):
        """Test premises"""
        premises = {
            'premises_name': 'INDUSTRIAL ESTATE',
            'sub_premises_type': 'UNIT',
            'sub_premises_number': 1,
            'sub_premises_suffix': '-3'
            }
        self.assertEqual(self.address.premises(), premises, "Incorrect Exception IV Unit premises")

class TestExceptionIVStall(unittest.TestCase):
    """Test Address Exception IV Stall"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'organisation_name': "QUIRKY CANDLES LTD",
            'building_name': "STALL 4",
            'thoroughfare_name': "MARKET",
            'thoroughfare_descriptor': "SQUARE",
            'post_town': "LIVERPOOL",
            'postcode': "L8 1LH"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = [
            "QUIRKY CANDLES LTD",
            "STALL 4",
            "MARKET SQUARE",
            "LIVERPOOL",
            "L8 1LH"
            ]
        self.assertEqual(
            self.address.as_list(), address, "Incorrect Exception IV Stall list format"
            )

    def test_string(self):
        """Test conversion to a string"""
        address = "QUIRKY CANDLES LTD, STALL 4, MARKET SQUARE, LIVERPOOL. L8 1LH"
        self.assertEqual(
            self.address.as_str(), address, "Incorrect Exception IV Stall string format"
            )

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = (
            "QUIRKY CANDLES LTD",
            "STALL 4",
            "MARKET SQUARE",
            "LIVERPOOL",
            "L8 1LH"
            )
        self.assertEqual(
            self.address.as_tuple(), address, "Incorrect Exception IV Stall tuple format"
            )

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "QUIRKY CANDLES LTD",
            'line_2': "STALL 4",
            'line_3': "MARKET SQUARE",
            'post_town': "LIVERPOOL",
            'postcode': "L8 1LH"
            }
        self.assertEqual(
            self.address.as_dict(), address, "Incorrect Exception IV Stall dict format"
            )

    def test_premises(self):
        """Test premises"""
        premises = {'premises_type': 'STALL', 'premises_number': 4}
        self.assertEqual(self.address.premises(), premises, "Incorrect Exception IV Stall premises")

class TestExceptionIVRearOf(unittest.TestCase):
    """Test Address Exception IV Rear Of"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'organisation_name': "FIONA'S FLOWERS",
            'building_name': "REAR OF 5A",
            'thoroughfare_name': "HIGH",
            'thoroughfare_descriptor': "STREET",
            'post_town': "GATESHEAD",
            'postcode': "NE8 1BH"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = [
            "FIONA'S FLOWERS",
            "REAR OF 5A",
            "HIGH STREET",
            "GATESHEAD",
            "NE8 1BH"
            ]
        self.assertEqual(
            self.address.as_list(), address, "Incorrect Exception IV RearOf list format"
            )

    def test_string(self):
        """Test conversion to a string"""
        address = "FIONA'S FLOWERS, REAR OF 5A, HIGH STREET, GATESHEAD. NE8 1BH"
        self.assertEqual(
            self.address.as_str(), address, "Incorrect Exception IV RearOf string format"
            )

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = (
            "FIONA'S FLOWERS",
            "REAR OF 5A",
            "HIGH STREET",
            "GATESHEAD",
            "NE8 1BH"
            )
        self.assertEqual(
            self.address.as_tuple(), address, "Incorrect Exception IV RearOf tuple format"
            )

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "FIONA'S FLOWERS",
            'line_2': "REAR OF 5A",
            'line_3': "HIGH STREET",
            'post_town': "GATESHEAD",
            'postcode': "NE8 1BH"
            }
        self.assertEqual(
            self.address.as_dict(), address, "Incorrect Exception IV RearOf dict format"
            )

    def test_premises(self):
        """Test premises"""
        premises = {'premises_type': 'REAR OF', 'premises_number': 5, 'premises_suffix': 'A'}
        self.assertEqual(
            self.address.premises(), premises, "Incorrect Exception IV RearOf premises"
            )

class TestExceptionIVBlock(unittest.TestCase):
    """Test Address Exception IV BLock"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'building_name': "BLOCK B",
            'thoroughfare_name': "WELLESLEY",
            'thoroughfare_descriptor': "ROAD",
            'post_town': "CROYDON",
            'postcode': "CR9 1AT"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = [
            "BLOCK B",
            "WELLESLEY ROAD",
            "CROYDON",
            "CR9 1AT"
            ]
        self.assertEqual(
            self.address.as_list(), address, "Incorrect Exception IV Block list format"
            )

    def test_string(self):
        """Test conversion to a string"""
        address = "BLOCK B, WELLESLEY ROAD, CROYDON. CR9 1AT"
        self.assertEqual(
            self.address.as_str(), address, "Incorrect Exception IV Block string format"
            )

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = (
            "BLOCK B",
            "WELLESLEY ROAD",
            "CROYDON",
            "CR9 1AT"
            )
        self.assertEqual(
            self.address.as_tuple(), address, "Incorrect Exception IV Block tuple format"
            )

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "BLOCK B",
            'line_2': "WELLESLEY ROAD",
            'post_town': "CROYDON",
            'postcode': "CR9 1AT"
            }
        self.assertEqual(
            self.address.as_dict(), address, "Incorrect Exception IV Block dict format"
            )

    def test_premises(self):
        """Test premises"""
        premises = {'premises_type': 'BLOCK', 'premises_suffix': 'B'}
        self.assertEqual(self.address.premises(), premises, "Incorrect Exception IV Block premises")

if __name__ == '__main__':
    unittest.main()



"""Thoroughfare and Locality Mixin"""

class ThoroughfareLocalityMixin():
    """Thoroughfare and locality processing"""

    @classmethod
    @property
    def dependent_thoroughfare_attrs(cls):
        """Returns Paf dependent thoroughfare elements"""
        return ('dependent_thoroughfare_name', 'dependent_thoroughfare_descriptor')

    @classmethod
    @property
    def thoroughfare_attrs(cls):
        """Returns Paf thoroughfare elements"""
        return ('thoroughfare_name', 'thoroughfare_descriptor')

    @classmethod
    @property
    def locality_attrs(cls):
        """Returns Paf locality elements"""
        return ('double_dependent_locality', 'dependent_locality')

    @classmethod
    @property
    def thoroughfare_and_locality_attrs(cls):
        """Returns derived thoroughfare and locality attributes"""
        return ('dependent_thoroughfare', 'thoroughfare') + cls.locality_attrs

    @property
    def thoroughfares_and_localities(self):
        """Returns thoroughfares and localities list"""
        attrs = self.thoroughfare_and_locality_attrs
        return [getattr(self, k) for k in attrs if not self.is_used_or_empty(k)]

    @property
    def dependent_thoroughfare(self):
        """Returns dependent thoroughfare"""
        return ' '.join(
            filter(None, [getattr(self, k, None) for k in self.dependent_thoroughfare_attrs])
            )

    @property
    def thoroughfare(self):
        """Returns thoroughfare"""
        return ' '.join(filter(None, [getattr(self, k, None) for k in self.thoroughfare_attrs]))

    @property
    def first_thoroughfare_or_locality_attr(self):
        """Returns name of first populated thoroughfare or locality attribute"""
        for k in self.thoroughfare_and_locality_attrs:
            if not self.is_empty(k):
                return k
        return None

    @property
    def first_thoroughfare_or_locality(self):
        """Returns first populated thoroughfare or locality value"""
        attr = self.first_thoroughfare_or_locality_attr
        return getattr(self, attr) if attr is not None else ''

    def is_used_or_empty(self, attr):
        """Returns if attribute value is empty or has already been used"""
        if self.is_empty(attr):
            return True
        return(
            attr == self.first_thoroughfare_or_locality_attr
                and self.premises.includes_first_thoroughfare_or_locality
            )



"""Premises Extender Mixin"""

import sys
from .rule000 import Rule000 # pylint: disable=unused-import
from .rule001 import Rule001 # pylint: disable=unused-import
from .rule010 import Rule010 # pylint: disable=unused-import
from .rule100 import Rule100 # pylint: disable=unused-import
from .rule011 import Rule011 # pylint: disable=unused-import
from .rule101 import Rule101 # pylint: disable=unused-import
from .rule110 import Rule110 # pylint: disable=unused-import
from .rule111 import Rule111 # pylint: disable=unused-import

class ExtenderMixin():
    """Dynamic Premises processing"""

    @property
    def rule(self):
        """Returns premises rule class"""
        rule = ''.join(['0' if self.is_empty(k) else '1' for k in self.building_attrs])
        return getattr(sys.modules[__name__], 'Rule' + rule)

    def extend(self):
        """Dynamically extends instance with appropriate rule"""
        base_cls = self.__class__
        object.__setattr__(self, '__class__', type(base_cls.__name__, (base_cls, self.rule), {}))



"""Building Types"""

from .split import SplitMixin

class BuildingTypeMixin(SplitMixin):
    """Determines if a string represents a known building type"""

    @classmethod
    @property
    def known_building_types(cls):
        """Returns known building types"""
        return (
            "BACK OF", "BLOCK", "BLOCKS", "BUILDING", "MAISONETTE", "MAISONETTES", "REAR OF",
            "SHOP", "SHOPS", "STALL", "STALLS", "SUITE", "SUITES", "UNIT", "UNITS", "PO BOX"
            )

    @classmethod
    @property
    def known_sub_building_types(cls):
        """Returns known sub-building types"""
        return cls.known_building_types + ("FLAT", "FLATS")

    def is_known_building_type(self, attr='building_name'):
        """Returns if attribute starts with a known type"""
        return self.but_last_word(attr) in self.known_building_types

    def is_known_sub_building_type(self, attr='sub_building_name'):
        """Returns if attribute starts with a known type"""
        return self.but_last_word(attr) in self.known_sub_building_types



"""Rule 6"""

from .premisable import PremisableMixin

class Rule110(PremisableMixin):
    """Rule 6 processing"""

    @property
    def rule_attrs(self):
        """Returns premises list"""
        if self.is_exception('sub_building_name'):
            return('sub_name_and_name',)
        if self.is_exception('building_name'):
            return('sub_building_name', 'name_and_thoroughfare_or_locality')
        return ('sub_building_name', 'building_name')

    @property
    def includes_first_thoroughfare_or_locality(self):
        """Returns if premises includes first thoroughfare or locality"""
        return (not self.is_exception('sub_building_name')) and self.is_exception('building_name')



"""Rule 3"""

from .dependent_premisable import DependentPremisableMixin

class Rule010(DependentPremisableMixin):
    """Rule 3 processing"""

    @property
    def rule_attrs(self):
        """Returns premises list"""
        if self.is_exception('building_name'):
            return ('name_and_thoroughfare_or_locality',)
        if self.is_split_exception('building_name'):
            return ('building_name_but_last_word', 'name_last_word_and_thoroughfare_or_locality')
        return ('building_name',)

    @property
    def includes_first_thoroughfare_or_locality(self):
        """Returns if premises includes first thoroughfare or locality"""
        return self.is_exception('building_name') or self.is_split_exception('building_name')



"""Splitting String Attributes"""

import re

class SplitMixin():
    """Methods to extract various parts of a attribute value"""

    @classmethod
    def __but_last_word(cls, string):
        """Returns all but last word of the string"""
        try:
            *first, _ = string.split()
            return ' '.join(first)
        except ValueError:
            return ''

    @classmethod
    def __last_word(cls, string):
        """Returns last word of the string"""
        try:
            *_, last = string.split()
            return last
        except ValueError:
            return string

    @classmethod
    def __leading_digits(cls, string):
        """Returns the leading digits from the string"""
        match = re.match(r'^\d+', string)
        if match:
            return int(match.group(0))
        return 0

    @classmethod
    def __after_leading_digits(cls, string):
        """Returns the characters after the leading digits from the string"""
        match = re.match(r'^\d+', string)
        if match:
            end_of_digits = match.end()
            return string[end_of_digits:]
        return ''

    def but_last_word(self, attr):
        """Returns all but last word of the attribute"""
        return self.__but_last_word(str(getattr(self, attr, '')))

    def last_word(self, attr):
        """Returns last word of the attribute"""
        return self.__last_word(str(getattr(self, attr, '')))

    def leading_digits(self, attr):
        """Returns the leading digits from the attribute"""
        return self.__leading_digits(str(getattr(self, attr, '')))

    def after_leading_digits(self, attr):
        """Returns the characters after the leading digits from the attribute"""
        return self.__after_leading_digits(str(getattr(self, attr, '')))



"""Attribute Mixin"""

class AttributeMixin():
    """Address elements and derived properties"""

    @classmethod
    @property
    def premises_attrs(cls):
        """Returns Paf premises elements"""
        return (
            'organisation_name', 'department_name',
                'sub_building_name', 'building_name', 'building_number',
                'po_box_number'
            )

    @classmethod
    @property
    def post_attrs(cls):
        """Returns Paf post elements"""
        return ('post_town', 'postcode')

    @classmethod
    @property
    def other_attrs(cls):
        """Returns Paf other elements"""
        return ('concatenation_indicator', 'udprn')

    @classmethod
    @property
    def attrs(cls):
        """Returns all Paf address elements"""
        return(
            cls.premises_attrs
                + cls.dependent_thoroughfare_attrs
                + cls.thoroughfare_attrs
                + cls.locality_attrs
                + cls.post_attrs
                + cls.other_attrs
            )

    def is_empty(self, attr):
        """Returns if attribute value is empty"""
        return getattr(self, attr, '') == ''



"""Rule 7"""

from .premisable import PremisableMixin

class Rule111(PremisableMixin):
    """Rule 7 processing"""

    @property
    def rule_attrs(self):
        """Returns premises list"""
        if self.is_zero('building_number'):
            return('sub_name_comma_name',)
        if self.is_exception('sub_building_name'):
            return('sub_name_and_name', 'number_and_thoroughfare_or_locality')
        return ('sub_building_name', 'building_name', 'number_and_thoroughfare_or_locality')

    @property
    def includes_first_thoroughfare_or_locality(self):
        """Returns if premises includes first thoroughfare or locality"""
        return not self.is_zero('building_number')

    def is_zero(self, attr):
        """Returns if attribute value is a 0"""
        return getattr(self, attr, '') == "0"



"""Test Address PO Box formatting"""

import unittest
import paf

class TestPoBox(unittest.TestCase):
    """Test Address PO Box"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'po_box_number': "61",
            'post_town': "FAREHAM",
            'postcode': "PO14 1UX"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["PO BOX 61", "FAREHAM", "PO14 1UX"]
        self.assertEqual(self.address.as_list(), address, "Incorrect PO Box list format")

    def test_string(self):
        """Test conversion to a string"""
        address = "PO BOX 61, FAREHAM. PO14 1UX"
        self.assertEqual(self.address.as_str(), address, "Incorrect PO Box string format")

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("PO BOX 61", "FAREHAM", "PO14 1UX")
        self.assertEqual(self.address.as_tuple(), address, "Incorrect PO Box tuple format")

    def test_dict(self):
        """Test conversion to a dict"""
        address = {'line_1': "PO BOX 61", 'post_town': "FAREHAM", 'postcode': "PO14 1UX"}
        self.assertEqual(self.address.as_dict(), address, "Incorrect PO Box dict format")

    def test_premises(self):
        """Test premises"""
        premises = {'premises_type': 'PO BOX', 'premises_number': 61}
        self.assertEqual(self.address.premises(), premises, "Incorrect PO Box premises")

if __name__ == '__main__':
    unittest.main()



"""Test Address Rule 5 formatting"""

import unittest
import paf

class TestRule5(unittest.TestCase):
    """Test Address Rule 5"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'sub_building_name': "FLAT 1",
            'building_number': "12",
            'thoroughfare_name': "LIME TREE",
            'thoroughfare_descriptor': "AVENUE",
            'post_town': "BRISTOL",
            'postcode': "BS8 4AB"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["FLAT 1", "12 LIME TREE AVENUE", "BRISTOL", "BS8 4AB"]
        self.assertEqual(self.address.as_list(), address, "Incorrect Rule 5 list format")

    def test_string(self):
        """Test conversion to a string"""
        address = "FLAT 1, 12 LIME TREE AVENUE, BRISTOL. BS8 4AB"
        self.assertEqual(self.address.as_str(), address, "Incorrect Rule 5 string format")

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("FLAT 1", "12 LIME TREE AVENUE", "BRISTOL", "BS8 4AB")
        self.assertEqual(self.address.as_tuple(), address, "Incorrect Rule 5 tuple format")

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "FLAT 1",
            'line_2': "12 LIME TREE AVENUE",
            'post_town': "BRISTOL",
            'postcode': "BS8 4AB"
            }
        self.assertEqual(self.address.as_dict(), address, "Incorrect Rule 5 dict format")

    def test_premises(self):
        """Test premises"""
        premises = {'premises_number': 12, 'sub_premises_type': 'FLAT', 'sub_premises_number': 1}
        self.assertEqual(self.address.premises(), premises, "Incorrect Rule 5 premises")

class TestRule5WithConcatenation(unittest.TestCase):
    """Test Address Rule 5 with Concatenation"""

    def setUp(self):
        """Set up Address instance"""
        self.address = paf.Address({
            'sub_building_name': "A",
            'building_number': "12",
            'thoroughfare_name': "HIGH",
            'thoroughfare_descriptor': "STREET NORTH",
            'dependent_locality': "COOMBE BISSETT",
            'post_town': "SALISBURY",
            'postcode': "SP5 4NA",
            'concatenation_indicator': "Y"
            })

    def test_list(self):
        """Test conversion to an list"""
        address = ["12A HIGH STREET NORTH", "COOMBE BISSETT", "SALISBURY", "SP5 4NA"]
        self.assertEqual(
            self.address.as_list(), address, "Incorrect Rule 5 with concatenate list format"
            )

    def test_string(self):
        """Test conversion to a string"""
        address = "12A HIGH STREET NORTH, COOMBE BISSETT, SALISBURY. SP5 4NA"
        self.assertEqual(
            self.address.as_str(), address, "Incorrect Rule 5 with concatenate string format"
            )

    def test_tuple(self):
        """Test conversion to a tuple"""
        address = ("12A HIGH STREET NORTH", "COOMBE BISSETT", "SALISBURY", "SP5 4NA")
        self.assertEqual(
            self.address.as_tuple(), address, "Incorrect Rule 5 with concatenate tuple format"
            )

    def test_dict(self):
        """Test conversion to a dict"""
        address = {
            'line_1': "12A HIGH STREET NORTH",
            'line_2': "COOMBE BISSETT",
            'post_town': "SALISBURY",
            'postcode': "SP5 4NA"
            }
        self.assertEqual(
            self.address.as_dict(), address, "Incorrect Rule 5 with concatenate dict format"
            )

    def test_premises(self):
        """Test premises"""
        premises = {'premises_number' : 12, 'sub_premises_name': 'A'}
        self.assertEqual(
            self.address.premises(), premises, "Incorrect Rule 5 with concatenate premises"
            )

if __name__ == '__main__':
    unittest.main()
